import { Utils } from "../utils/utils.js";
import { TxosAggregatorResult } from "./txos-aggregator-result.js";
class ComputeLinkMatrixTask {
    constructor(idxIl, il, ir, dOut) {
        this.idxIl = idxIl;
        this.il = il;
        this.ir = ir;
        this.dOut = dOut;
    }
}
export class TxosAggregator {
    constructor(logger) {
        this.Logger = logger;
    }
    matchAggByVal(allAgg, fees, intraFees) {
        const allInAggVal = allAgg.inAgg.allAggVal;
        const allOutAggVal = allAgg.outAgg.allAggVal;
        const allUniqueInAggVal = [...new Set(allInAggVal)].sort((a, b) => a - b);
        const allUniqueOutAggVal = [...new Set(allOutAggVal)].sort((a, b) => a - b);
        const allMatchInAgg = [];
        const matchInAggToVal = new Map();
        const valToMatchOutAgg = new Map();
        const hasIntraFees = intraFees?.hasFees ?? false;
        const feesTaker = hasIntraFees ? fees + intraFees.feesTaker : 0;
        const feesMaker = hasIntraFees ? -intraFees.feesMaker : 0;
        const PROGRESS_ID = "matchAggByVal";
        for (let i = 0; i < allUniqueInAggVal.length; i++) {
            const inAggVal = allUniqueInAggVal[i];
            Utils.logProgress(this.Logger, PROGRESS_ID, i, allUniqueInAggVal.length);
            for (const outAggVal of allUniqueOutAggVal) {
                const diff = inAggVal - outAggVal;
                if (!hasIntraFees && diff < 0) {
                    break;
                }
                else {
                    const condNoIntrafees = !hasIntraFees && diff <= fees;
                    const condIntraFees = hasIntraFees &&
                        ((diff <= 0 && diff >= feesMaker) ||
                            (diff >= 0 && diff <= feesTaker));
                    if (condNoIntrafees || condIntraFees) {
                        for (const [inIdx, element] of allInAggVal.entries()) {
                            if (element === inAggVal && !allMatchInAgg.includes(inIdx)) {
                                allMatchInAgg.push(inIdx);
                                matchInAggToVal.set(inIdx, inAggVal);
                            }
                        }
                        if (!valToMatchOutAgg.has(inAggVal)) {
                            valToMatchOutAgg.set(inAggVal, []);
                        }
                        const keysMatchOutAgg = valToMatchOutAgg.get(inAggVal);
                        for (const [idx, element] of allOutAggVal.entries()) {
                            if (element === outAggVal) {
                                keysMatchOutAgg.push(idx);
                            }
                        }
                    }
                }
            }
        }
        Utils.logProgressDone(this.Logger, PROGRESS_ID, allUniqueInAggVal.length);
        return {
            allMatchInAgg,
            matchInAggToVal,
            valToMatchOutAgg,
        };
    }
    computeInAggCmbn(aggMatches) {
        const aggs = [...aggMatches.allMatchInAgg];
        aggs.shift();
        const mat = new Map();
        if (aggs.length > 0) {
            const tgt = aggs.pop() || 0;
            const PROGRESS_ID = "computeInAggCmbn";
            for (let i = 0; i <= tgt; i++) {
                if (aggs.includes(i)) {
                    const jMax = Math.min(i, tgt - i + 1);
                    for (let j = 0; j < jMax; j++) {
                        if ((i & j) === 0 && aggs.includes(j)) {
                            const aggChilds = this.createLine(mat, i + j);
                            aggChilds.push([i, j]);
                        }
                    }
                }
                Utils.logProgress(this.Logger, PROGRESS_ID, i, tgt, `${mat.size} matches`);
            }
            Utils.logProgressDone(this.Logger, PROGRESS_ID, tgt, `${mat.size} matches`);
        }
        return mat;
    }
    createLine(mat, i) {
        let line = mat.get(i);
        if (!line) {
            line = [];
            mat.set(i, line);
        }
        return line;
    }
    checkDtrmLinks(txos, allAgg, aggMatches) {
        const nbIns = txos.inputs.size;
        const nbOuts = txos.outputs.size;
        const matCmbn = Array.from(new Array(nbOuts), () => new Array(nbIns).fill(0));
        const inCmbn = Array.from({ length: nbIns }).fill(0);
        for (const inEntry of aggMatches.matchInAggToVal.entries()) {
            const inIdx = inEntry[0];
            const val = inEntry[1];
            const outIdxArr = aggMatches.valToMatchOutAgg.get(val);
            for (const outIdx of outIdxArr) {
                this.updateLinkCmbn(matCmbn, inIdx, outIdx, allAgg);
                const inIndexes = allAgg.inAgg.allAggIndexes[inIdx];
                for (const inIndex of inIndexes) {
                    const currentValue = inCmbn[inIndex];
                    inCmbn[inIndex] = currentValue + 1;
                }
            }
        }
        const nbCmbn = inCmbn[0];
        return this.findDtrmLinks(matCmbn, nbCmbn);
    }
    computeLinkMatrix(txos, allAgg, aggMatches, matInAggCmbn, maxDuration) {
        let nbTxCmbn = 0;
        const itGt = 2 ** txos.inputs.size - 1;
        const otGt = 2 ** txos.outputs.size - 1;
        const dLinks = new Map();
        const dOutInitial = new Map();
        const dOutEntry = new Map();
        dOutEntry.set(0, [1, 0]);
        dOutInitial.set(otGt, dOutEntry);
        const stack = [];
        const rootTask = new ComputeLinkMatrixTask(0, 0, itGt, dOutInitial);
        stack.push(rootTask);
        const rootIrcs = matInAggCmbn.get(rootTask.ir);
        const rootLenIrcs = rootIrcs ? rootIrcs.length : 0;
        const startTime = Date.now();
        let totalIterations = 0;
        let iterations = 0;
        const PROGRESS_ID = "computeLinkMatrix";
        while (stack.length > 0) {
            const currTime = Date.now();
            const deltaTimeSeconds = (currTime - startTime) / 1000;
            if (maxDuration && deltaTimeSeconds >= maxDuration) {
                this.Logger.logInfo("maxDuration limit reached!");
                return new TxosAggregatorResult(0, []);
            }
            let t = stack.at(-1);
            let nIdxIl = t.idxIl;
            const ircs = matInAggCmbn.get(t.ir);
            const lenIrcs = ircs ? ircs.length : 0;
            for (let i = t.idxIl; i < lenIrcs; i++) {
                iterations++;
                nIdxIl = i;
                const nIl = ircs[i][1];
                if (nIl > t.il) {
                    const nIr = ircs[i][0];
                    Utils.logProgress(this.Logger, PROGRESS_ID, rootTask.idxIl, rootLenIrcs, `${iterations}/${totalIterations}, ${dLinks.size} dlinks`);
                    const ndOut = this.runTask(nIl, nIr, aggMatches, otGt, t.dOut);
                    t.idxIl = i + 1;
                    iterations++;
                    stack.push(new ComputeLinkMatrixTask(0, nIl, nIr, ndOut));
                    const newIrcs = matInAggCmbn.get(nIr);
                    const newLenIrcs = newIrcs ? newIrcs.length : 0;
                    totalIterations += newLenIrcs;
                    break;
                }
                else {
                    nIdxIl = ircs.length;
                    break;
                }
            }
            if (nIdxIl > lenIrcs - 1) {
                t = stack.pop();
                if (stack.length === 0) {
                    nbTxCmbn = t.dOut.get(otGt).get(0)[1];
                }
                else {
                    const pt = stack.at(-1);
                    this.onTaskCompleted(t, pt, dLinks);
                }
            }
        }
        Utils.logProgressDone(this.Logger, PROGRESS_ID, rootLenIrcs, `${dLinks.size} dlinks`);
        return this.finalizeLinkMatrix(allAgg, itGt, otGt, dLinks, nbTxCmbn);
    }
    finalizeLinkMatrix(allAgg, itGt, otGt, dLinks, nbTxCmbn) {
        const linksClear = this.newLinkCmbn(allAgg);
        const links = JSON.parse(JSON.stringify(linksClear));
        this.updateLinkCmbn(links, itGt, otGt, allAgg);
        nbTxCmbn++;
        const linksX = links.length;
        const linksY = links[0].length;
        const PROGRESS_ID = "finalizeLinkMatrix";
        let i = 0;
        for (const firstKeyEntry of dLinks.entries()) {
            const key0 = firstKeyEntry[0];
            Utils.logProgress(this.Logger, PROGRESS_ID, i++, dLinks.size, `Processing dLink... ${firstKeyEntry[1].size} x (${linksX}x${linksY})`);
            for (const secondKeyEntry of firstKeyEntry[1].entries()) {
                const key1 = secondKeyEntry[0];
                const mult = secondKeyEntry[1];
                const linkCmbn = JSON.parse(JSON.stringify(linksClear));
                this.updateLinkCmbn(linkCmbn, key0, key1, allAgg);
                for (let i1 = 0; i1 < linksX; i1++) {
                    for (let j = 0; j < linksY; j++) {
                        links[i1][j] = links[i1][j] + linkCmbn[i1][j] * mult;
                    }
                }
            }
            i++;
        }
        Utils.logProgressDone(this.Logger, PROGRESS_ID, dLinks.size);
        return new TxosAggregatorResult(nbTxCmbn, links);
    }
    onTaskCompleted(t, pt, dLinks) {
        const il = t.il;
        const ir = t.ir;
        for (const doutEntry of t.dOut.entries()) {
            const or = doutEntry[0];
            const lOl = doutEntry[1];
            const rKey = [ir, or];
            for (const olEntry of lOl.entries()) {
                const ol = olEntry[0];
                const nbPrnt = olEntry[1][0];
                const nbChld = olEntry[1][1];
                const lKey = [il, ol];
                const nbOccur = nbChld + 1;
                this.addDLinkLine(rKey, nbPrnt, dLinks);
                this.addDLinkLine(lKey, nbPrnt * nbOccur, dLinks);
                const pOr = ol + or;
                const plOl = pt.dOut.get(pOr);
                for (const plOlEntry of plOl.entries()) {
                    plOlEntry[1][1] += nbOccur;
                }
            }
        }
    }
    addDLinkLine(key, addValue, dLinks) {
        let subMap = dLinks.get(key[0]);
        if (!subMap) {
            subMap = new Map();
            dLinks.set(key[0], subMap);
        }
        const currentValue = subMap.get(key[1]) || 0;
        subMap.set(key[1], currentValue + addValue);
    }
    runTask(nIl, nIr, aggMatches, otGt, dOut) {
        const ndOut = new Map();
        for (const oREntry of dOut.entries()) {
            const oR = oREntry[0];
            const sol = otGt - oR;
            let nbPrt = 0;
            for (const value of oREntry[1].values()) {
                nbPrt += value[0];
            }
            const valIl = aggMatches.matchInAggToVal.get(nIl);
            for (const nOl of aggMatches.valToMatchOutAgg.get(valIl)) {
                if ((sol & nOl) === 0) {
                    const nSol = sol + nOl;
                    const nOr = otGt - nSol;
                    const valIr = aggMatches.matchInAggToVal.get(nIr);
                    const matchOutAgg = aggMatches.valToMatchOutAgg.get(valIr);
                    if ((nSol & nOr) === 0 && matchOutAgg.includes(nOr)) {
                        const ndOutVal = this.ndOutLine(ndOut, nOr);
                        ndOutVal.set(nOl, [nbPrt, 0]);
                    }
                }
            }
        }
        return ndOut;
    }
    ndOutLine(ndOut, idx) {
        let ndOutVal = ndOut.get(idx);
        if (!ndOutVal) {
            ndOutVal = new Map();
            ndOut.set(idx, ndOutVal);
        }
        return ndOutVal;
    }
    newLinkCmbn(allAgg) {
        let maxOutIndex = 0;
        for (const indexes of allAgg.outAgg.allAggIndexes) {
            const max = Math.max(...indexes);
            if (max > maxOutIndex) {
                maxOutIndex = max;
            }
        }
        let maxInIndex = 0;
        for (const indexes of allAgg.inAgg.allAggIndexes) {
            const max = Math.max(...indexes);
            if (max > maxInIndex) {
                maxInIndex = max;
            }
        }
        return Array.from({ length: maxOutIndex + 1 })
            .fill(null)
            .map(() => new Array(maxInIndex + 1).fill(0));
    }
    updateLinkCmbn(matCmbn, inAgg, outAgg, allAgg) {
        const outIndexes = allAgg.outAgg.allAggIndexes[outAgg];
        const inIndexes = allAgg.inAgg.allAggIndexes[inAgg];
        for (const inIndex of inIndexes) {
            for (const outIndex of outIndexes) {
                const value = matCmbn[outIndex][inIndex];
                matCmbn[outIndex][inIndex] = value + 1;
            }
        }
        return matCmbn;
    }
    findDtrmLinks(matCmbn, nbCmbn) {
        const dtrmCoords = new Set();
        for (const [i, element] of matCmbn.entries()) {
            for (const [j, element_] of element.entries()) {
                if (element_ === nbCmbn) {
                    dtrmCoords.add([i, j]);
                }
            }
        }
        return dtrmCoords;
    }
}
//# sourceMappingURL=txos-aggregator.js.map