import { BoltzmannResult, } from "./beans/boltzmann-result.js";
import { createBoltzmannSettings, } from "./beans/boltzmann-settings.js";
import { TxProcessor } from "./processor/tx-processor.js";
import { CustomMap, Logger } from "./utils/utils.js";
export class Boltzmann {
    constructor(options) {
        this.settings = createBoltzmannSettings(options);
        this.Logger = Logger(this.settings.logLevel);
        this.txProcessor = new TxProcessor(this.settings.maxDuration, this.settings.maxTxos, this.Logger);
    }
    process(txos) {
        return this.processWithOptions(txos, this.settings.maxCjIntrafeesRatio, this.settings.linkerOptions);
    }
    processWithOptions(txos, maxCjIntrafeesRatio, linkerOptions) {
        const t1 = Date.now();
        this.Logger.logDebug("SETTINGS: ", JSON.stringify(this.settings));
        const txosObject = {
            inputs: new CustomMap(txos.inputs),
            outputs: new CustomMap(txos.outputs),
        };
        const txProcessorResult = this.txProcessor.processTx(txosObject, maxCjIntrafeesRatio, linkerOptions);
        const duration = (Date.now() - t1) / 1000;
        return new BoltzmannResult(duration, txProcessorResult);
    }
}
//# sourceMappingURL=boltzmann.js.map